# generate checksum for a list of files

from pathlib import Path
from tempfile import NamedTemporaryFile

import click
from fabric import Connection
from invoke import run
from tqdm import tqdm

from .common import split_target
from .exceptions import ChecksumCompareFailed, ChecksumGenerationFailed
from .watcher import LineWatcher


def src_checksum(src, dst, hash, output_filename, ascii, width, count):
    return _checksum(True, src, dst, hash, output_filename, ascii, width, count)


def dst_checksum(src, dst, hash, output_filename, ascii, width, count):
    return _checksum(False, src, dst, hash, output_filename, ascii, width, count)


# noqa: C901


def _checksum(is_src, src, dst, hash, output_filename, ascii, width, count):
    """generate BSD-style checksum for each file in target, returning local file containing result"""

    if is_src:
        host, target = split_target(src)
        mode = "source"
    else:
        host, target = split_target(dst)
        _, src_target = split_target(src)
        # target = str(Path(target) / Path(src_target).stem)
        mode = "destination"

    if host:
        label = f"{host}:{target}"
    else:
        label = target
        target = Path(target).resolve()

    click.echo(f"Generating {mode} checksums for {label}")

    if host:
        runner = Connection(host).run
    else:
        runner = run

    hproc = runner(f"which 2>/dev/null {hash} || which {hash}sum", hide=True, warn=True)
    if hproc.ok:
        hasher = hproc.stdout.strip()
    else:
        breakpoint()
        pass
    if hasher.endswith("sum"):
        hasher += " --tag"

    options = "--progress"
    if width:
        options += f" --width {width}"
    if ascii:
        options += " --ascii"

    with NamedTemporaryFile("w+", delete_on_close=False) as tempfile:
        with tqdm(total=count, unit=" lines", miniters=1, delay=1, ncols=width, ascii=ascii) as bar:

            def line_callback(line):
                bar.update(1)

            genproc = runner(
                f"cd {target}; find . -type f -exec {hasher} \x5c\x7b\x5c\x7d \x5c;",
                warn=True,
                watchers=[LineWatcher(None, None, line_callback)],
                hide=True,
                out_stream=tempfile.file,
            )

            if genproc.failed:
                raise ChecksumGenerationFailed(genproc.stderr)

        tempfile.close()
        click.echo(f"Sorting {mode} checksums...")
        with Path(output_filename).open("w") as ofp:
            run(f"sort {tempfile.name}", out_stream=ofp, hide=True)

    return Path(output_filename)


def compare_checksums(src_sums, dst_sums):
    """take two pathlike args, returning True if their contents are identical"""

    result = run(f"diff {str(src_sums)} {str(dst_sums)}", hide=True, warn=True)
    if result.failed:
        Path(".checksums.src").write_text(Path(src_sums).read_text())
        Path(".checksums.dst").write_text(Path(dst_sums).read_text())
        Path(".checksums.out").write_text(result.stdout)
        Path(".checksums.err").write_text(result.stderr)
        raise ChecksumCompareFailed("details written to .checksums.*")

    return result.ok



# utility functions


def split_target(target: str) -> (str, str):
    """return (hostname, path) for target; hostname is None for local target"""
    host, _, path = str(target).partition(":")
    if path:
        return host, path
    else:
        return None, host


def parse_int(field):
    return int(field.replace(",", ""))



# checksum test cases

import subprocess
import sys

import pytest
from invoke import run

from cptree.checksum import dst_checksum, src_checksum
from cptree.cptree import cptree

HASH = "sha256"


@pytest.fixture
def output(test_datadir):
    return test_datadir / "output"


@pytest.fixture
def file_list(output):
    return output / "file_list"


@pytest.fixture
def reference_checksums(output):
    return output / f"reference.{HASH}"


@pytest.fixture
def output_filename(output):
    return str(output / f"test.{HASH}")


@pytest.fixture
def count():
    def _count(src_dir):
        ret = int(subprocess.check_output(f"find {src_dir} -type f | wc -l", shell=True, text=True))
        return ret

    return _count


@pytest.fixture
def compare_checksums(output):
    def _compare_checksums(test_sums):
        reference_sums = (output / "reference").with_suffix(test_sums.suffix)
        cmd = f"diff -y {str(reference_sums)} {str(test_sums)}"
        result = run(cmd, hide=True, warn=True)
        if result.failed:
            print(result.stdout, flush=True)
            print(result.stderr, file=sys.stderr, flush=True)
        return result.ok

    return _compare_checksums


def test_checksum_local_src(local_src, local_dst, output, compare_checksums, count):
    test_sums = src_checksum(local_src, local_dst, HASH, output / f"local_src.{HASH}", False, 120, count(local_src))
    assert compare_checksums(test_sums)


def test_checksum_local_dst(local_src, local_dst, output, compare_checksums, count):
    assert cptree(str(local_src) + "/", local_dst, delete="force-no-countdown", hash=None) == 0
    test_sums = dst_checksum(local_src, local_dst, HASH, output / f"local_dst.{HASH}", False, 120, count(local_src))
    assert compare_checksums(test_sums)


def test_checksum_remote_src(remote_src, local_src, local_dst, output, compare_checksums, count):
    assert cptree(str(local_src) + "/", remote_src, delete="force-no-countdown", hash=None) == 0
    test_sums = src_checksum(remote_src, local_dst, HASH, output / f"remote_src.{HASH}", False, 120, count(local_src))
    assert compare_checksums(test_sums)


def test_checksum_remote_dst(local_src, remote_dst, output, compare_checksums, count):
    assert cptree(str(local_src) + "/", remote_dst, delete="force-no-countdown", hash=None) == 0
    test_sums = dst_checksum(local_src, remote_dst, HASH, output / f"remote_dst.{HASH}", False, 120, count(local_src))
    assert compare_checksums(test_sums)



# cptree exceptions


class Fail(Exception):
    pass


class InvalidDirectory(Fail):
    pass


class ChecksumCompareFailed(Fail):
    pass


class RsyncTransferFailed(Fail):
    pass


class ChecksumGenerationFailed(Fail):
    pass



# cptree implementation

import re
import shlex
import shutil
import sys
from pathlib import Path
from tempfile import TemporaryDirectory

import click
from invoke import run
from tqdm import tqdm

from .checksum import compare_checksums, dst_checksum, src_checksum
from .common import parse_int
from .exceptions import ChecksumCompareFailed, RsyncTransferFailed
from .verify import verify_dst_directory, verify_output_directory, verify_src_directory
from .watcher import LineWatcher


def mkopts(kwargs):
    opts = " "
    for k, v in kwargs:
        if v:
            opts.append(f"{k} {v} ")
        else:
            opts.append(f"{k} ")
    opts = opts.strip()
    return opts


def prescan(src, dst, opts):
    click.echo("Scanning...\r", nl=False)
    scanproc = run(f"rsync -a --list-only {opts} {src} {dst}", hide=True, warn=True)
    if scanproc.ok:
        items = scanproc.stdout.strip().split("\n")
    else:
        click.echo(scanproc.stderr, err=True)
        sys.exit(scanproc.return_code)
    counts = {
        "d": 0,
        "-": 0,
        "l": 0,
        "p": 0,
        "c": 0,
        "b": 0,
    }
    names = {
        "d": "directories",
        "-": "files",
        "l": "links",
        "p": "pipes",
        "c": "char_devs",
        "b": "block_devs",
    }

    pattern = re.compile(r"^(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s(.*)$")
    size = 0
    files = []
    for item in items:
        fields = pattern.match(item).groups()
        length = parse_int(fields[1])
        size += int(length)
        file_type = item[0]
        counts[file_type] += 1
        if file_type == "-":
            files.append(fields[4])

    details = []
    for k, v in counts.items():
        if v:
            details.append(f"{names[k]}={v}")

    click.echo(f"Transferring {len(items)} items: [{', '.join(details)}]")

    if (counts["b"] + counts["c"]) > 0:
        click.confirm(
            "Attempting transfer of device nodes.  Are you certain you know what you're doing?",
            abort=True,
        )

    return size, len(items), files


def cptree(*args, output_dir=None, **kwargs):
    """call _cptree with work_dir from argument or a temp dir"""

    if output_dir:
        kwargs["output_dir"] = Path(output_dir)
        return _cptree(*args, **kwargs)
    else:
        with TemporaryDirectory() as temp_dir:
            kwargs["output_dir"] = Path(temp_dir)
            return _cptree(*args, **kwargs)


def _verify_dirs(src, dst, output_dir, create, delete):
    verify_output_directory(output_dir)
    verify_src_directory(src)
    verify_dst_directory(dst, create, delete)


def _rsync_echo(cmd):
    words = shlex.split(cmd)
    words.remove("-ii")
    ipos = words.index("--info")
    words.pop(ipos)
    words.pop(ipos)
    fpos = words.index("--out-format")
    words.pop(fpos)
    words.pop(fpos)
    return " ".join(words)


def _cptree(  # noqa: C901
    src,
    dst,
    *,
    create=None,
    delete=None,
    progress=True,
    output_dir=None,
    hash=None,
    rsync=True,
    rsync_args=None,
):
    _verify_dirs(src, dst, output_dir, create, delete)

    if rsync_args is None:
        rsync_args = ""

    bytes, items, files = prescan(src, dst, rsync_args)

    file_list = output_dir / "file_list"
    file_list.write_text("\n".join(["./" + f for f in files]) + "\n")

    ascii = bool(progress == "ascii")
    width = shutil.get_terminal_size().columns - 1

    bar = tqdm(
        total=bytes,
        unit="B",
        unit_scale=True,
        miniters=1,
        ncols=width,
        delay=1,
        ascii=ascii,
        disable=bool(progress in ["disable", False, None]),
    )

    def line_callback(line):
        click.echo(line)

    def file_callback(filename, count, length, codes):
        bar.set_description(f"[{count}/{items}]", refresh=False)

    def progress_callback(chunk, progress):
        bar.update(chunk)

    if progress:
        progress_opts = "-ii --info PROGRESS2 --out-format '~%i %l %f'"
        watcher = LineWatcher(file_callback, progress_callback, None)
    else:
        progress_opts = ""
        watcher = LineWatcher(None, None, line_callback)

    if rsync:
        cmd = f"rsync -avz {progress_opts} {rsync_args} {src} {dst}"

        # echo the rsync command without progress data output options
        click.echo(_rsync_echo(cmd))

        proc = run(
            cmd,
            watchers=[watcher],
            in_stream=None,
            hide=True,
            warn=True,
        )
        rsync_stderr = proc.stderr.strip().split("\n")
    else:
        click.echo("Skipping rsync transfer")
        rsync_stderr = []
        proc = None

    bar.close()

    for error in rsync_stderr:
        if error.strip():
            click.echo(error, err=True)

    if proc is not None and proc.failed:
        raise RsyncTransferFailed(f"rsync failed with error code: {proc.return_code}")

    if hash:
        _verify_hashes(src, dst, hash, output_dir, ascii, width, len(files))

    if proc is None:
        return 0
    else:
        return proc.return_code


def _verify_hashes(src, dst, hash, output_dir, ascii, width, count):
    src_sums = src_checksum(src, dst, hash, output_dir / f"src.{hash}", ascii, width, count)
    dst_sums = dst_checksum(src, dst, hash, output_dir / f"dst.{hash}", ascii, width, count)
    compare_checksums(src_sums, dst_sums)
    with src_sums.open("r") as sfp:
        ssums = len(sfp.readlines())
    with dst_sums.open("r") as dfp:
        dsums = len(dfp.readlines())
    if ssums == dsums:
        click.echo(f"\nSuccessful Transfer. Verified matching {hash.upper()} hashes on {ssums} files.\n")
    else:
        raise ChecksumCompareFailed(f"checksum count mismatch: src={ssums} dst={dsums}")
